/* <hack>
   This is just a huge ugly hack to get things compiled.
   Hopefully will disappear quickly, once we get everything
   to compile... */
	.text
	.align	8
	.globl	_sigpause_common
_sigpause_common:	retl;nop
	.globl	breakpoint
breakpoint:	retl;nop
	.globl	do_cee
do_cee:	retl;nop
	.globl	do_cee_tl1
do_cee_tl1:	retl;nop
	.globl	do_dae
do_dae:	retl;nop
	.globl	do_dae_tl1
do_dae_tl1:	retl;nop
	.globl	do_div0
do_div0:	retl;nop
	.globl	do_div0_tl1
do_div0_tl1:	retl;nop
	.globl	do_fpdis
do_fpdis:	retl;nop
	.globl	do_fpdis_tl1
do_fpdis_tl1:	retl;nop
	.globl	do_fpieee
do_fpieee:	retl;nop
	.globl	do_fpieee_tl1
do_fpieee_tl1:	retl;nop
	.globl	do_fpother
do_fpother:	retl;nop
	.globl	do_fpother_tl1
do_fpother_tl1:	retl;nop
	.globl	do_iae
do_iae:	retl;nop
	.globl	do_iae_tl1
do_iae_tl1:	retl;nop
	.globl	do_ill
do_ill:	retl;nop
	.globl	do_ill_tl1
do_ill_tl1:	retl;nop
	.globl	do_irq
do_irq:	retl;nop
	.globl	do_irq_tl1
do_irq_tl1:	retl;nop
	.globl	do_lddfmna
do_lddfmna:	retl;nop
	.globl	do_lddfmna_tl1
do_lddfmna_tl1:	retl;nop
	.globl	do_mna
do_mna:	retl;nop
	.globl	do_mna_tl1
do_mna_tl1:	retl;nop
	.globl	do_paw
do_paw:	retl;nop
	.globl	do_paw_tl1
do_paw_tl1:	retl;nop
	.globl	do_privact
do_privact:	retl;nop
	.globl	do_privop
do_privop:	retl;nop
	.globl	do_signal
do_signal:	retl;nop
	.globl	do_stdfmna
do_stdfmna:	retl;nop
	.globl	do_stdfmna_tl1
do_stdfmna_tl1:	retl;nop
	.globl	do_tof
do_tof:	retl;nop
	.globl	do_tof_tl1
do_tof_tl1:	retl;nop
	.globl	do_vaw
do_vaw:	retl;nop
	.globl	do_vaw_tl1
do_vaw_tl1:	retl;nop
	.globl	floppy_hardint
floppy_hardint:	retl;nop
	.globl	get_cpuid
get_cpuid:	retl;nop
	.globl	getcc
getcc:	retl;nop
	.globl	halt
halt:	retl;nop
	.globl	indirect_syscall
indirect_syscall:	retl;nop
	.globl	install_linux_ticker
install_linux_ticker:	retl;nop
	.globl	install_obp_ticker
install_obp_ticker:	retl;nop
	.globl	linux_dbvec
linux_dbvec:	retl;nop
	.globl	linux_num_cpus
linux_num_cpus:	retl;nop
	.globl	netbsd_syscall
netbsd_syscall:	retl;nop
	.globl	setcc
setcc:	retl;nop
	.globl	solaris_syscall
solaris_syscall:	retl;nop
	.globl	sunos_mmap
sunos_mmap:	retl;nop
	.globl	sunos_syscall
sunos_syscall:	retl;nop
	.globl	svr4_getcontext
svr4_getcontext:	retl;nop
	.globl	svr4_setcontext
svr4_setcontext:	retl;nop
	.globl	sunos_accept
sunos_accept:	retl;nop
	.globl	sunos_audit
sunos_audit:	retl;nop
	.globl	sunos_brk
sunos_brk:	retl;nop
	.globl	sunos_execv
sunos_execv:	retl;nop
	.globl	sunos_fpathconf
sunos_fpathconf:	retl;nop
	.globl	sunos_getdents
sunos_getdents:	retl;nop
	.globl	sunos_getdirentries
sunos_getdirentries:	retl;nop
	.globl	sunos_getdomainname
sunos_getdomainname:	retl;nop
	.globl	sunos_getdtablesize
sunos_getdtablesize:	retl;nop
	.globl	sunos_getgid
sunos_getgid:	retl;nop
	.globl	sunos_gethostid
sunos_gethostid:	retl;nop
	.globl	sunos_getpid
sunos_getpid:	retl;nop
	.globl	sunos_getsockopt
sunos_getsockopt:	retl;nop
	.globl	sunos_getuid
sunos_getuid:	retl;nop
	.globl	sunos_indir
sunos_indir:	retl;nop
	.globl	sunos_ioctl
sunos_ioctl:	retl;nop
	.globl	sunos_killpg
sunos_killpg:	retl;nop
	.globl	sunos_madvise
sunos_madvise:	retl;nop
	.globl	sunos_mctl
sunos_mctl:	retl;nop
	.globl	sunos_mincore
sunos_mincore:	retl;nop
	.globl	sunos_mount
sunos_mount:	retl;nop
	.globl	sunos_nop
sunos_nop:	retl;nop
	.globl	sunos_nosys
sunos_nosys:	retl;nop
	.globl	sunos_open
sunos_open:	retl;nop
	.globl	sunos_pathconf
sunos_pathconf:	retl;nop
	.globl	sunos_poll
sunos_poll:	retl;nop
	.globl	sunos_read
sunos_read:	retl;nop
	.globl	sunos_readv
sunos_readv:	retl;nop
	.globl	sunos_recv
sunos_recv:	retl;nop
	.globl	sunos_sbrk
sunos_sbrk:	retl;nop
	.globl	sunos_select
sunos_select:	retl;nop
	.globl	sunos_semsys
sunos_semsys:	retl;nop
	.globl	sunos_send
sunos_send:	retl;nop
	.globl	sunos_setpgrp
sunos_setpgrp:	retl;nop
	.globl	sunos_setsockopt
sunos_setsockopt:	retl;nop
	.globl	sunos_shmsys
sunos_shmsys:	retl;nop
	.globl	sunos_sigaction
sunos_sigaction:	retl;nop
	.globl	sunos_sigblock
sunos_sigblock:	retl;nop
	.globl	sunos_sigsetmask
sunos_sigsetmask:	retl;nop
	.globl	sunos_sstk
sunos_sstk:	retl;nop
	.globl	sunos_sysconf
sunos_sysconf:	retl;nop
	.globl	sunos_uname
sunos_uname:	retl;nop
	.globl	sunos_vadvise
sunos_vadvise:	retl;nop
	.globl	sunos_wait4
sunos_wait4:	retl;nop
	.globl	sunos_write
sunos_write:	retl;nop
	.globl	sunos_writev
sunos_writev:	retl;nop
	.globl	sys_ptrace
sys_ptrace:	retl;nop
	.globl	sys_sigreturn
sys_sigreturn:	retl;nop
	.globl	sys_sigstack
sys_sigstack:	retl;nop
	.globl	sys_sigsuspend
sys_sigsuspend:	retl;nop
	.globl	syscall_trace
syscall_trace:	retl;nop
	.globl	sys32_ptrace
sys32_ptrace:	retl;nop
	.globl	do_sigpause
do_sigpause:	retl;nop
